/*
* This script applies or removes a status to all component versions mentioned in the latest snapshot created in / applied to an environment
* or from a named snapshot.
* This is to make is easier to make sure that component versions have the right statuses to pass environment gates.
* 
* (c) Copyright IBM Corporation 2020. All Rights Reserved.
*
*/

import com.urbancode.ud.client.UDRestClient;
import com.urbancode.commons.web.util.PercentCodec;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.client.methods.HttpUriRequest;
import com.urbancode.commons.util.IO;
import com.urbancode.air.AirPluginTool;

//	General Concept
//	===============
//	1. If not supplied, Find the most recent snapshot applied (or created in ) the environment
//	2. Get the complete list of component versions (this might include several incremental versions of a component)
//	3. Apply a status to each of these component versions
//	4. Log the versions the status was applied to


// 	Get the UCD Plugin framework class to retrieve properties
	final AirPluginTool apt = new AirPluginTool(new File(args[0]), new File(args[1]));
	final Properties props = apt.getStepProperties();
	def env = System.getenv()

	// Values retrieved from UCD plugin framework
	UCDURL=env['AH_WEB_URL']
	UCDUSER=apt.getAuthTokenUsername()
	UCDPASSWD=apt.getAuthToken()
	removeStatus = props['removeStatus'] == "true"
	versionStatusName=props['versionStatusName']
	applicationName=props['applicationName']
	environmentName=props['environmentName']
	snapshotName=props['snapshotName']

        // Now lets create an HTTP client using the UCD REST API Client and start pulling stuff back from the server
        myHttpClient= UDRestClient.createHttpClient(UCDUSER, UCDPASSWD);
	if ( snapshotName == "" )
	{
		println "Operating on the latest snapshot of environment '${environmentName}' of application '${applicationName}'"
		snapshot = getLatestSnapshot(applicationName, environmentName)
                if ( snapshot == "")
                {
                        println "ERROR: There is no defined latest snapshot for this environment. A snapshot must be explicitly specified."
                        System.exit(1)
                }
	}
	else
	{
		println "Operating on the snapshot '${snapshotName}' of application '${applicationName}'"
		snapshot = getSnapshot(applicationName, snapshotName)
	}
 
	snid = snapshot.id
	String snapshotName = snapshot.name

	if (!removeStatus) 
		println "Applying Statues to component versions belonging to snapshot '${snapshotName}'"	
	else
		println "Removing Statues from component versions belonging to snapshot '${snapshotName}'"

	// Get the list of components 
	verCount = 0
	snapVer=getSnapshotVersions(snid)
	snapVer.each
	{
		// Iterate over desired versions for each component
		it.desiredVersions.each
		{
			// Apply or remove status
			verCount++
			if (! removeStatus )
			{
				println "Adding Status: ${versionStatusName} to version: ${it.name} of component: ${it.component.name}"
				applyStatusToVersion(versionStatusName, it.id )
			}
			else
			{
				println "Removing Status: ${versionStatusName} from version: ${it.name} of component: ${it.component.name}"
				removeStatusFromVersion(versionStatusName, it.id)
			}
		}
	}

	if (! removeStatus)
		println "Status Applied to ${verCount} Versions."
	else
		println "Status Removed from ${verCount} Versions."

	// Exit OK
	System.exit(0)
//
// Utility functions to implement API calls.
//
def getLatestSnapshot(String application, String environment)
{

        String uri = this.UCDURL + "/cli/environment/${sanitizePathSegment(environment)}/snapshot?application=${sanitizePathSegment(application)}"
        return restApiGet(uri)
}

def getSnapshot(String application, String snapshot)
{
        String uri = this.UCDURL + "/cli/snapshot/getSnapshot?application=${sanitizePathSegment(application)}&snapshot=${sanitizePathSegment(snapshot)}"
        return restApiGet(uri)
}


def getSnapshotVersions(String snapshotID)
{
	// Get the list of versions in the desired snapshot
        String uri = this.UCDURL + "/cli/snapshot/getSnapshotVersions?snapshot=${snapshotID}"
        return restApiGet(uri)
}
def removeStatusFromVersion(String statusName, String versionID)
{
        String uri = this.UCDURL + "/cli/version/status?version=${versionID}&status=${sanitizePathSegment(statusName)}"
        return restApiDelete(uri)
}
def applyStatusToVersion(String statusName, String versionID)
{
        String uri = this.UCDURL + "/cli/version/addStatus?version=${versionID}&status=${sanitizePathSegment(statusName)}"
        return restApiPut(uri)
}
def restApiPut(String uri)
{
	final HttpPut method = new HttpPut(uri);
	final CloseableHttpResponse response = this.invokeMethod(method);
}

def restApiGet(String uri)
{
        final HttpGet method = new HttpGet(uri);
        final CloseableHttpResponse response = invokeMethod((HttpRequestBase)method);
        final String body =  this.getBody(response);
//  	if there is no body, return an empty string not null
        if (body == null)
                return ""

        def slurper = new groovy.json.JsonSlurper()
        json = slurper.parseText( body )
        return json
}
def restApiDelete(String uri)
{
        final HttpDelete method = new HttpDelete(uri);
        final CloseableHttpResponse response = invokeMethod((HttpRequestBase)method);
}

String sanitizePathSegment(final String path)
{
	PercentCodec encoder = new PercentCodec();
	return encoder.encode(path);
}

CloseableHttpResponse invokeMethod(final HttpRequestBase request)
{
        final CloseableHttpResponse response = myHttpClient.execute((HttpUriRequest)request);
        final int status = response.getStatusLine().getStatusCode();
        if (status > 299) {
                throw new IOException(String.format("%d %s\n%s", response.getStatusLine().getStatusCode(), response.getStatusLine().getReasonPhrase(), this.getBody(response)));
        }
        return response;
}

protected String getBody(final CloseableHttpResponse response) throws IOException 
{
	String result = null;
        final StringBuilder builder = new StringBuilder();
        if (response.getStatusLine().getStatusCode() != 204) 
	{
        	final InputStream body = response.getEntity().getContent();
                if (body != null) 
		{
                	final Reader reader = IO.reader(body, IO.utf8());
                        try {
                        	IO.copy(reader, (Appendable)builder);
                            }
                            finally {
                                reader.close();
                            }
                }
                result = builder.toString();
        }
        return result;
}
